/////////////////////////////////////////////////////
//
//    Wii Gyro
//
//    by kako
//
/////////////////////////////////////////////////////


#include <afxwin.h>
#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h> 

#define FUNC __declspec(dllimport) __stdcall
extern "C" HANDLE FUNC OpenHidHandle(unsigned short vendor_id, unsigned short product_id);
extern "C" void FUNC ReadReport(HANDLE handle,unsigned char *InputReport,int *len);
extern "C" void FUNC WriteReport(HANDLE handle,unsigned char *OutputReport, int *len);
extern "C" void FUNC CloseHidHandle(HANDLE handle);

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg,WPARAM wParam, LPARAM lParam);

HWND hWnd;
HPEN pen;
HPEN penR;
HPEN penG;
HPEN penB;
HBRUSH brush;
HFONT font;

#define I_BLACK RGB(0,0,0)
#define I_RED RGB(200,0,0)
#define I_GREEN RGB(0,200,0)
#define I_BLUE RGB(0,0,200)
#define I_WHITE RGB(255,255,255)

HANDLE hRsDevHandle;

const USHORT VID = 0x057e; // Nintendo (?)
const USHORT PID = 0x0306; // Wii-Remote (?)

#define MAXREPORTSIZE 256
int		InputLength,OutputLength;
unsigned char	InputReport[MAXREPORTSIZE];
unsigned char	OutputReport[MAXREPORTSIZE];

unsigned char	button1,button2;
unsigned char	Ax,Ay,Az;
unsigned char	wii_remote_loop_flag;
unsigned char	wii_remote_in_use;
unsigned char	Err;
int				Rot;

int				Ix1,Iy1,Ix2,Iy2;
int				Ix3,Iy3,Ix4,Iy4;

void Wii_Remote_Input(void);
void Wii_Remote_mode_0x31(void);
int Wii_Remote_check(void);
void Display_Event(void);

void Sleep2(int msec)
{
	LARGE_INTEGER f;
	LARGE_INTEGER t1,t2;
	double t;
	QueryPerformanceFrequency(&f);
	QueryPerformanceCounter(&t1);
	while (1) {
		QueryPerformanceCounter(&t2);
		t = (double)t2.QuadPart - (double)t1.QuadPart;
		t /= (double)f.QuadPart;
		t *= 1000;
		if (t>(double)msec) { return; }
	}
}

void Report_0x11(unsigned char data)
{
	OutputReport[0]=0x11; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x12(unsigned char data1,unsigned char data2)
{
	OutputReport[0]=0x12; // Report ID
	OutputReport[1]=data1;
	OutputReport[2]=data2;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x13(unsigned char data)
{
	OutputReport[0]=0x13; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x17(unsigned char adr,unsigned char len,unsigned char d1)
{
	OutputReport[0]=0x17; // Report ID  ... Auto Read == dump(?)
	OutputReport[1]=0x04; // 00(EEPROM) or 04(other ... 1byten)
	OutputReport[2]=adr; // I2C ADRS
	OutputReport[3]=0x00; // ignore?
	OutputReport[4]=d1;  // mem adr .. ADRSɑ΂ăR}hs, length16oCgȏŉZ
	OutputReport[5]=0x00; // length H
	OutputReport[6]=len;  // length L
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Report_0x1a(unsigned char data)
{
	OutputReport[0]=0x1a; // Report ID
	OutputReport[1]=data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Wii_Remote_mode_0x30(void)
{
	// Set Report Type
	Report_0x12(0x00,0x30);  // button ONLY Report Mode
}

void Wii_Remote_mode_0x31(void)
{
	// Set Report Type
	Report_0x12(0x00,0x31);  // Acc Sensor + button  Report Mode
}

void Wii_Remote_Input(void)
{
	unsigned char r;
	int s;
	ReadReport( hRsDevHandle , InputReport, &InputLength);
	
	r = InputReport[0];
	if ( (r & 0xF0) == 0x30 )  {
		button1=InputReport[1]; button2=InputReport[2];
		
		if ( r & 0x01 ) {
			Ax = InputReport[3];
			Ay = InputReport[4];
			Az = InputReport[5];
		}
		if ( r & 0x02 ) {
			// ir
			Ix1 = InputReport[6];
			Iy1 = InputReport[7];
			s   = InputReport[8];
			Ix1 += (s & 0x30) <<4;
			Iy1 += (s & 0xC0) <<2;
			
			Ix2 = InputReport[9];
			Iy2 = InputReport[10];
			s   = InputReport[11];
			Ix2 += (s & 0x30) <<4;
			Iy2 += (s & 0xC0) <<2;
			
			Ix3 = InputReport[12];
			Iy3 = InputReport[13];
			s   = InputReport[14];
			Ix3 += (s & 0x30) <<4;
			Iy3 += (s & 0xC0) <<2;
			
			Ix4 = InputReport[15];
			Iy4 = InputReport[16];
			s   = InputReport[17];
			Ix4 += (s & 0x30) <<4;
			Iy4 += (s & 0xC0) <<2;
		}
	}
}

#define WT 1

void Gyro_read(unsigned char *r1 , unsigned char *r2)
{
	int d1,d2,d3,d4,d5;
	Report_0x17(0x92, 2, 0x90 /* Data-read */);
	while (1) {
		ReadReport( hRsDevHandle , InputReport, &InputLength);
		if ( InputReport[0]==0x31 ) {
			button1=InputReport[1]; button2=InputReport[2];
			Ax = InputReport[3];
			Ay = InputReport[4];
			Az = InputReport[5];
		}
		if (InputReport[0]==0x21) { break; }
	}
	d1=InputReport[3]; // result ssss-ffff  ... ssss receive-size / ffff==0x07 error
	d2=InputReport[4]; d3=InputReport[5];  // data-offset
	d4=InputReport[6]; d5=InputReport[7]; // data
	Err = d1 & 0x0F;
	*r1 = d4;
	*r2 = d5;
}


void Wii_Remote_Input_Main(void *)
{
	// wii remote input thread
	while (wii_remote_loop_flag) {
		wii_remote_in_use=1;
		Wii_Remote_Input();
		Sleep(100);
	}
	wii_remote_in_use=0;
	return;
}

uintptr_t wii_th;

int Wii_Remote_check(void)
{
	Ax=0;
	wii_remote_loop_flag=1;
	wii_th = _beginthread( Wii_Remote_Input_Main , 0 , NULL );
	Sleep2(100);
	wii_remote_loop_flag=0;
	return (Ax!=0);
}

void Stop_Wii_remote_th(void)
{
	if (wii_remote_in_use==1) {
		TerminateThread( (HANDLE)wii_th , -1); // kill thread  if it stopped
	}
}

void Wii_Gyro_main(void *)
{
	unsigned char d1,d2;
	int d;
	
	Sleep2(500);

	wii_remote_loop_flag=1;
	
	while (wii_remote_loop_flag) {
		Wii_Remote_mode_0x31();
		Gyro_read(&d1, &d2);
		d=d1*256+d2;
		if (d>0x8000) {
			Rot = -(0xFFFF - d);
		} else {
			Rot = d;
		}
		if (button2==0x80) { // Home button
			SendMessage(hWnd,WM_DESTROY,0,0); // quit
		}
		Display_Event();
	}
}


int APIENTRY WinMain(HINSTANCE hInst,HINSTANCE,LPSTR cmd,int show)
{
	MSG msg;
	HACCEL hAccel;
	WNDCLASSEX wndclass;
	HANDLE hMutex;
	
	hMutex = CreateMutex(NULL, TRUE, "wii-test-001");
	if (!hMutex) { return FALSE; }
	if (GetLastError() == ERROR_ALREADY_EXISTS) {
		return FALSE;
	}
	
	hRsDevHandle = OpenHidHandle(VID,PID);
	if ( hRsDevHandle == INVALID_HANDLE_VALUE) {
		MessageBox(0,"Wii-Remote not found.",0,0);
		return 1;
	}
	
	Wii_Remote_mode_0x31();
	if ( !Wii_Remote_check() ) {
		MessageBox(0,"Wii-Remote cannot open.",0,0);
		Stop_Wii_remote_th();
		CloseHidHandle( hRsDevHandle );
		return 1;
	}
	
	Report_0x11(0xF0); // LED1 on
	_beginthread( Wii_Gyro_main , 0 , NULL );
	
	wndclass.cbSize        = sizeof(wndclass);
	wndclass.style         = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc   = WndProc;
	wndclass.cbClsExtra    = 0;
	wndclass.cbWndExtra    = 0;
	wndclass.hInstance     = hInst;
	wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
	wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName  = NULL;
	wndclass.lpszClassName = "wnd-001";
	wndclass.hIconSm       = 0;
	RegisterClassEx (&wndclass);
	hWnd=CreateWindow("wnd-001", "Wii Remote Gyro Sensor Test",
			WS_OVERLAPPEDWINDOW,10,10,520,420,NULL,NULL,hInst,NULL);
	if (hWnd==NULL) { return FALSE; }
	ShowWindow(hWnd,show); UpdateWindow(hWnd);
	while (GetMessage(&msg,NULL,0,0)) {
		TranslateMessage(&msg); DispatchMessage(&msg);
	}
	return msg.wParam;
}

void Display_Event(void)
{
	//InvalidateRect(hWnd,NULL,FALSE); //display update 1
	InvalidateRect(hWnd,NULL,TRUE); //display update 2
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg,WPARAM wParam, LPARAM lParam)
{
	char buf[255];
	switch (msg) {
	 case WM_CREATE:
		{
			pen=CreatePen(PS_SOLID,6,I_BLACK);
			penR=CreatePen(PS_SOLID,5,I_RED);
			penG=CreatePen(PS_SOLID,5,I_GREEN);
			penB=CreatePen(PS_SOLID,5,I_BLUE);
			brush=CreateSolidBrush(I_WHITE);
			font = CreateFont(  16, 0, 0, 0,FW_DONTCARE,false,false,false,
					DEFAULT_CHARSET,false,false,false,false,"System");
		}
		break;
	 case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc;
			hdc=BeginPaint(hWnd,&ps);
			SelectObject(hdc,pen);
			SelectObject(hdc,brush);
			SelectObject(hdc,font);
			{
				int x1,y1;
				int x2,y2;
				TextOut(hdc,200, 30,"X-axis Acceleration  ",20);
				TextOut(hdc,200, 80,"Y-axis Acceleration  ",20);
				TextOut(hdc,200,130,"Z-axis Acceleration  ",20);
				TextOut(hdc,200,230,"Gyro Rotation        ",20);
				SelectObject(hdc,penR);
				x1 = 250;  x2 = ((int)Ax - 128)*5 + x1; y1 =50; y2=y1;
				MoveToEx(hdc,x1,y1,NULL); LineTo(hdc,x2,y2);
				SelectObject(hdc,penG);
				x1 = 250;  x2 = ((int)Ay - 128)*5 + x1; y1 =100; y2=y1;
				MoveToEx(hdc,x1,y1,NULL); LineTo(hdc,x2,y2);
				SelectObject(hdc,penB);
				x1 = 250;  x2 = ((int)Az - 128)*5 + x1; y1 =150; y2=y1;
				MoveToEx(hdc,x1,y1,NULL); LineTo(hdc,x2,y2);
				SelectObject(hdc,pen);
				x1 = 250;  x2 = Rot + x1; y1 =250; y2=y1;
				MoveToEx(hdc,x1,y1,NULL); LineTo(hdc,x2,y2);
			}
			EndPaint(hWnd,&ps);
		}
		break;
	 case WM_KEYDOWN:
		if (wParam==0x1B) { SendMessage(hWnd,WM_DESTROY,0,0); } // ESCL[
		break;
	 case WM_DESTROY:
		DeleteObject(pen); DeleteObject(brush); DeleteObject(font);
		PostQuitMessage(0);
		wii_remote_loop_flag=0;
		Wii_Remote_mode_0x30();
		Report_0x11(0x10); // LED1 on
		CloseHidHandle( hRsDevHandle );
		break;
	 default:
		return DefWindowProc(hWnd,msg,wParam,lParam);
	}
	return 0;
}
